import os
import sys
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error

class TestEvents(FakeChrootTestCase):

    def test_nochange(self):
        self.fixture.check_apply("""
            resources:
              - Directory:
                  name: /etc/wibble
            """)

        rv = self.fixture.apply("""
            resources:
              - Directory:
                  name: /etc/wibble

              - File:
                  name: /frob/somedir/foo
                  policy:
                     apply:
                         when: apply
                         on: Directory[/etc/wibble]
            """)
        self.assertEqual(rv, error.NothingChanged.returncode)

    def test_recover(self):
        rv = self.fixture.apply("""
            resources:
              - Directory:
                  name: /etc/somedir

              - Directory:
                  name: /frob/somedir

              - File:
                  name: /frob/somedir/foo
                  policy:
                     apply:
                         when: apply
                         on: Directory[/etc/somedir]
            """)
        self.assertEqual(rv, error.PathComponentMissing.returncode)

        self.fixture.check_apply("""
            resources:
              - Directory:
                  name: /etc/somedir

              - Directory:
                  name: /frob

              - Directory:
                  name: /frob/somedir

              - File:
                  name: /frob/somedir/foo
                  policy:
                     apply:
                         when: apply
                         on: Directory[/etc/somedir]

            """,
               "--resume")

        self.failUnlessExists("/frob/somedir/foo")





"""

Classes that represent errors within yaybu.

What is listed here are the exceptions raised within Python, with an
explanation of their meaning. If you wish to detect a specific error on
invocation, you can do so via the return code of the yaybu process.

All yaybu errors have a returncode, which is returned from the yaybu program
if these errors occur. This is primarily for the test harness, but feel free
to rely on these, they should be stable.

A returncode of less than 128 is an error from within the range specified in
the errno library, which contains the standard C error codes.

These may have been actually returned from a shell command, or they may be
based on our interpretation of the failure mode they represent. Resources will
define the errors they may return. """

import errno

class Error(Exception):
    """ Base class for all yaybu specific exceptions. """
    returncode = 253

    def __init__(self, msg=""):
        self.msg = msg

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, self.msg)

class ParseError(Error):
    """ Root of exceptions that are caused by an error in input. """

    returncode = 128
    """ returns error code 128 to the invoking environment. """

class BindingError(Error):
    """ An error during policy binding. """

    returncode = 129
    """ returns error code 129 to the invoking environment. """

class ExecutionError(Error):
    """ Root of exceptions that are caused by execution failing in an unexpected way. """
    returncode = 130
    """ returns error code 130 to the invoking environment. """

class DpkgError(ExecutionError):
    """ dpkg returned something other than 0 or 1 """
    returncode = 131
    """ returns error code 131 to the invoking environment. """

class AptError(ExecutionError):
    """ An apt command failed unrecoverably. """
    returncode = 132
    """ returns error code 132 to the invoking environment. """

class CommandError(ExecutionError):
    """ A command from the Execute provider did not return the expected return
    code. """
    returncode = 133
    """ returns error code 133 to the invoking environment. """

class NoValidPolicy(ParseError):
    """ There is no valid policy for the resource. """

    returncode = 135
    """ returns error code 135 to the invoking environment. """

class NonConformingPolicy(ParseError):
    """ A policy has been specified, or has been chosen by default, but the
    parameters provided for the resource do not match those required for the
    policy. Check the documentation to ensure you have provided all required
    parameters. """
    returncode = 136
    """ returns error code 136 to the invoking environment. """

class NoSuitableProviders(ParseError):
    """ There are no suitable providers available for the policy and resource
    chosen. This may be because a provider has not been written for this
    Operating System or service, or it may be that you have not specified the
    parameters correctly. """
    returncode = 137
    """ returns error code 137 to the invoking environment. """

class TooManyProviders(ParseError):
    """ More than one provider matches the specified resource, and Yaybu is unable to choose between them. """
    returncode = 138
    """ returns error code 138 to the invoking environment. """

class InvalidProvider(ExecutionError):
    """ A provider is not valid. This is detected before any changes have been
    applied. """
    returncode = 139
    """ returns error code 139 to the invoking environment. """

class InvalidGroup(ExecutionError):
    """ The specified user group does not exist. """
    returncode = 140
    """ returns error code 140 to the invoking environment. """

class InvalidUser(ExecutionError):
    """ The specified user does not exist. """
    returncode = 141
    """ returns error code 141 to the invoking environment. """

class OperationFailed(ExecutionError):
    """ A general failure of an operation. For example, we tried to create a
    symlink, everything appeared to work but then a link does not exist. This
    should probably never happen. """
    returncode = 142
    """ returns error code 142 to the invoking environment. """

class BinaryMissing(ExecutionError):
    """ A specific error for an expected binary (ln, rm, etc.) not being
    present where expected. """
    returncode = 143
    """ returns error code 143 to the invoking environment. """

class DanglingSymlink(ExecutionError):
    """ The destination of a symbolic link does not exist. """
    returncode = 144
    """ returns error code 144 to the invoking environment. """

class UserAddError(ExecutionError):
    """ An error from the useradd command. It has a bunch of error codes of
    it's own. """
    returncode = 145
    """ returns error code 145 to the invoking environment. """

class PathComponentMissing(ExecutionError):
    """ A component of the path is not present """
    returncode = 146
    """ returns error code 146 to the invoking environment. """

class PathComponentNotDirectory(ExecutionError):
    """ A component of the path is in fact not a directory """
    returncode = 147
    """ returns error code 147 to the invoking environment. """

class SavedEventsAndNoInstruction(Error):
    """ There is a saved events file and the user has not decided what to do
    about it. Invoke with --resume or --no-resume. """
    returncode = 148
    """ returns error code 148 to the invoking environment. """

class MissingAsset(ExecutionError):
    """ An asset referenced by a resource could not be found on the Yaybu
    search path. """
    returncode = 149
    """ returns error code 149 to the invoking environment. """

class CheckoutError(Error):
    """ An insurmountable problem was encountered during checkout """
    returncode = 150
    """ returns error code 150 to the invoking environment. """

class Incompatible(Error):
    """ An incompatibility was detected and execution can't continue """
    returncode = 151

class MissingDependency(ExecutionError):
    """ A dependency required for a feature or provider is missing """
    returncode = 152

class NothingChanged(ExecutionError):
    """ Not really an error, but we need to know if this happens for our
    tests. This exception is never really raised, but it's useful to keep the
    error code here!"""
    returncode = 254
    """ returns error code 254 to the invoking environment. """

class ConnectionError(ExecutionError):
    """ An error occured while establishing a remote connection """
    returncode = 255

class SystemError(ExecutionError):
    """ An error represented by something in the errno list. """

    def __init__(self, returncode):
        # if the returncode is not in errno, this will blow up.
        self.msg = errno.errorcode[returncode]
        self.returncode = returncode




# coding=utf-8

import unittest
import datetime
from yaybu.core import argument
from yaybu.core import resource

class TestArguments(unittest.TestCase):

    def test_octal(self):
        class R_test_octal(resource.Resource):
            a = argument.Octal()
        r = R_test_octal(name="test")
        r.a = "666"
        self.assertEqual(r.a, 438)
        r.a = 666
        self.assertEqual(r.a, 438)

    def test_string(self):
        class R_test_string(resource.Resource):
            a = argument.String()
        r = R_test_string(name="test")
        r.a = "foo"
        self.assertEqual(r.a, "foo")
        r.a = u"foo"
        self.assertEqual(r.a, "foo")
        r.a = u"£40"
        self.assertEqual(r.a, u"£40")
        r.a = u"£40".encode("utf-8")
        self.assertEqual(r.a, u"£40")

    def test_integer(self):
        class R_test_integer(resource.Resource):
            a = argument.Integer()
        r = R_test_integer(name="test")
        r.a = 10
        self.assertEqual(r.a, 10)
        r.a = "10"
        self.assertEqual(r.a, 10)
        r.a = 10.5
        self.assertEqual(r.a, 10)

    def test_datetime(self):
        class R_test_datetime(resource.Resource):
            a = argument.DateTime()
        r = R_test_datetime(name="test")
        r.a = "2011-02-20"
        self.assertEqual(r.a, datetime.datetime(2011, 02, 20))





import os
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error

def sibpath(filename):
    return os.path.join(os.path.dirname(__file__), filename)


class TestLink(FakeChrootTestCase):

    def test_create_link(self):
        self.fixture.check_apply("""
            resources:
              - Link:
                  name: /etc/somelink
                  to: /etc
                  owner: root
                  group: root
            """)

        self.failUnlessExists("/etc/somelink")

    def test_remove_link(self):
        self.fixture.symlink("/", "/etc/toremovelink")
        rv = self.fixture.check_apply("""
            resources:
              - Link:
                  name: /etc/toremovelink
                  policy: remove
            """)
        self.failIfExists("/etc/toremovelink")

    def test_already_exists(self):
        self.fixture.symlink("/", "/etc/existing")
        rv = self.fixture.apply("""
            resources:
              - Link:
                  name: /etc/existing
                  to: /
        """)
        self.assertEqual(rv, 254)
        self.failUnlessEqual(self.fixture.readlink("/etc/existing"), "/")

    def test_already_exists_notalink(self):
        """ Test for the path already existing but is not a link. """
        with self.fixture.open("/bar_notalink", "w") as fp:
            fp.write("")
        with self.fixture.open("/foo", "w") as fp:
            fp.write("")

        self.fixture.check_apply("""
            resources:
                - Link:
                    name: /bar_notalink
                    to: /foo
            """)

        self.failUnlessEqual(self.fixture.readlink("/bar_notalink"), "/foo")

    def test_already_exists_pointing_elsewhere(self):
        """ Test for the path already existing but being a link to somewhere else. """
        self.fixture.touch("/baz")
        self.fixture.touch("/foo")
        self.fixture.symlink("/baz", "/bar_elsewhere")
        self.fixture.check_apply("""
            resources:
                - Link:
                    name: /bar_elsewhere
                    to: /foo
            """)
        self.failUnlessEqual(self.fixture.readlink("/bar_elsewhere"), "/foo")

    def test_dangling(self):
        rv = self.fixture.apply("""
        resources:
             - Link:
                 name: /etc/test_dangling
                 to: /etc/not_there
        """)
        self.assertEqual(rv, error.DanglingSymlink.returncode)

    def test_unicode(self):
        self.fixture.check_apply(open(sibpath("link_unicode1.yay")).read())




import os
import sys
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error

class TestAuditLog(FakeChrootTestCase):

    def test_auditlog_apply(self):
        self.fixture.check_apply("""
            resources:
                - File:
                    name: /test_auditlog_apply
            """)

        self.failUnlessExists("/var/log/yaybu.log")

    def test_auditlog_simulate(self):
        self.fixture.check_apply_simulate("""
            resources:
                - File:
                    name: /test_auditlog_simulate
            """)

        self.failIfExists("/var/log/yaybu.log")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import grp

from yaybu.core import provider
from yaybu.core import error
from yaybu import resources

import logging

logger = logging.getLogger("provider")

class Group(provider.Provider):

    policies = (resources.group.GroupApplyPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(Group, self).isvalid(*args, **kwargs)

    def get_group_info(self):
        fields = ("name", "passwd", "gid", "members",)

        try:
            info_tuple = grp.getgrnam(self.resource.name.encode("utf-8"))
        except KeyError:
            info = dict((f, None) for f in fields)
            info["exists"] = False
            return info

        info = {"exists": True}
        for i, field in enumerate(fields):
            info[field] = info_tuple[i]

        return info

    def apply(self, context):
        changed = False
        info = self.get_group_info()

        if info["exists"]:
            command = ["groupmod"]
        else:
            command = ["groupadd"]
            changed = True

        if self.resource.gid and info["gid"] != self.resource.gid:
            command.extend(["--gid", str(self.resource.gid)])

        command.extend([self.resource.name])

        if not changed:
            return False

        returncode, stdout, stderr = context.shell.execute(command)
        if returncode != 0:
            raise error.GroupError("%s failed with return code %d" % (self.resource, returncode))

        return True

class GroupRemove(provider.Provider):

    policies = (resources.group.GroupRemovePolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(GroupRemove, self).isvalid(*args, **kwargs)

    def apply(self, context):
        try:
            existing = grp.getgrnam(self.resource.name.encode("utf-8"))
        except KeyError:
            # If we get a key errror then there is no such group. This is good.
            return False

        command = ["groupdel", self.resource.name]

        returncode, stdout, stderr = context.shell.execute(command)
        if returncode != 0:
            raise error.GroupError("Removing group %s failed with return code %d" % (self.resource, returncode))

        return True




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaybu.core import provider
from yaybu.core import error
from yaybu import resources

class Prompt(provider.Provider):

    policies = (resources.prompt.PromptPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(Prompt, self).isvalid(*args, **kwargs)

    def apply(self, shell):

        raw_input(self.resource.question)



import os, shutil, grp

from yaybu.harness import FakeChrootTestCase
from yaybu.util import sibpath


class TestGroup(FakeChrootTestCase):

    def test_simple_group(self):
        self.fixture.check_apply("""
            resources:
                - Group:
                    name: test
            """)

        self.failUnless(self.fixture.get_group("test"))

    def test_group_with_gid(self):
        self.fixture.check_apply("""
            resources:
                - Group:
                    name: test
                    gid: 1111
            """)

        self.failUnless(self.fixture.get_group("test"))

    def test_existing_group(self):
        """ Test creating a group whose name already exists. """

        self.failUnless(self.fixture.get_group("users"))

        rv = self.fixture.apply("""
            resources:
                - Group:
                    name: users
            """)

        self.failUnlessEqual(rv, 254)

        self.failUnless(self.fixture.get_group("users"))

    def test_existing_gid(self):
        """ Test creating a group whose specified gid already exists. """
        rv = self.fixture.apply("""
            resources:
                - Group:
                    name: test
                    gid: 100
            """)

        self.failUnlessEqual(rv, 4)
        self.failUnlessRaises(KeyError, self.fixture.get_group, "test")

    def test_add_group_and_use_it(self):
        self.fixture.check_apply("""
            resources:
                - Group:
                    name: test
                - File:
                    name: /etc/test
                    group: test
                - Execute:
                    name: test-group
                    command: python -c "import os, grp; open('/etc/test2', 'w').write(grp.getgrgid(os.getgid()).gr_name)"
                    creates: /etc/test2
                    group: test
            """)
        self.failUnlessEqual(self.fixture.open("/etc/test2").read(), "test")


class TestGroupRemove(FakeChrootTestCase):

    def test_remove_existing(self):
        self.failUnless(self.fixture.get_group("users"))

        self.fixture.check_apply("""
            resources:
                - Group:
                    name: users
                    policy: remove
            """)

        self.failUnlessRaises(KeyError, self.fixture.get_group, "users")

    def test_remove_non_existing(self):
        self.failUnlessRaises(KeyError, self.fixture.get_group, "zzidontexistzz")

        rv = self.fixture.apply("""
            resources:
                - Group:
                    name: zzidontexistzz
                    policy: remove
            """)

        self.failUnlessEqual(rv, 254)

        self.failUnlessRaises(KeyError, self.fixture.get_group, "zzidontexistzz")





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import namedtuple

from yaybu.core import provider
from yaybu.providers.service import utils
from yaybu import resources

UpstartInfo = namedtuple('UpstartInfo', ['name', 'goal', 'status'])

class _UpstartServiceMixin(utils._ServiceMixin):

    features = ["restart", ]

    def _parse_line(self, line):
        """
        Converts a line of information from /sbin/status into a tuple of (name,
        goal, status). Any PID information is discared as we don't need it.

        Has to deal with strings like::

            hwclock stop/waiting
            ssh start/running, process 4931
            network-interface-security (network-manager) start/running
        """
        name, status = line.split(" ", 1)
        if status.startswith("("):
            instance, status = status.split(") ", 1)
            instance = instance.lstrip("(")
        else:
           instance = name
        goal, status = status.split("/", 1)
        if "," in status:
            status, _ = status.split(",", 1)
        return UpstartInfo(instance, goal, status)

    def _parse_status_output(self, statusblob):
        """
        Yields status information from the output of /sbin/status
        
        Can cope with multiple instances of a job like this::

            network-interface-security (network-manager) start/running
            network-interface-security (network-interface/eth0) start/running
            network-interface-security (network-interface/lo) start/running
            network-interface-security (networking) start/running
        """
        for line in statusblob.strip().split("\n"):
            if not line.strip():
                continue
            yield self._parse_line(line)

    def status(self, context):
        rv, stdout, stderr = context.shell.execute(["/sbin/status", self.resource.name], exceptions=False, passthru=True)
        if rv != 0:
            raise error.CommandError("Got exit code of %d whilst trying to determine status" % rv)

        if "Unknown job" in stderr:
            raise error.CommandError("Upstart does not know about this job")

        statuses = list(self._parse_status_output(stdout))

        if len(statuses) == 0:
            raise error.CommandError("Upstart returned no information for the job")

        if len(statuses) > 1:
            raise error.CommandError("The job has multiple statuses. This is currently not supported in Yaybu recipes.")

        try:
            return dict(start="running", stop="not-running")[statuses[0].goal]
        except KeyError:
            raise error.CommandError("The job has an unexpected goal of '%s'" % statuses[0].goal)

    @classmethod
    def isvalid(cls, policy, resource, yay):
        if not super(_UpstartServiceMixin, cls).isvalid(policy, resource, yay):
            return False
        if getattr(resource, policy.name):
            return False
        return os.path.exists("/sbin/start") and os.path.exists("/etc/init/%s.conf" % resource.name)

    def get_command(self, action):
        return ["/sbin/" + action, self.resource.name]


class Start(_UpstartServiceMixin, utils._Start, provider.Provider):
    pass


class Stop(_UpstartServiceMixin, utils._Stop, provider.Provider):
    pass


class Restart(_UpstartServiceMixin, utils._Restart, provider.Provider):
    pass





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, logging
import re

from yaybu.core.provider import Provider
from yaybu.core.error import CheckoutError, MissingDependency
from yaybu import resources

from yaybu.providers.filesystem.files import AttributeChanger


class Rsync(Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    REMOTE_NAME = "origin"

    @classmethod
    def isvalid(self, policy, resource, yay):
        return resource.scm and resource.scm.lower() == "rsync"

    def _get_svn_ignore(self, context, path):
        command = ["svn", "status", "--non-interactive", "--no-ignore", path]
        returncode, stdout, stderr = context.shell.execute(command, passthru=True)

        if not returncode == 0:
            raise CheckoutError("Could not generate updated .rsync-exclude for Subversion checkout")

        ignore = []
        for line in stdout.split("\n"):
            if not line.startswith("I"):
                continue
            ignore.append(os.path.relpath(line.lstrip("I").strip(), start=path))

        return ignore

    def _get_git_ignore(self, context, path):
        command = ["git", "clean", "-nXd"]
        returncode, stdout, stderr = context.shell.execute(command, cwd=path, passthru=True)

        if not returncode == 0:
            raise CheckoutError("Could not generate updated .rsync-exclude for Git checkout")

        ignore = []
        for line in stdout.split("\n"):
            ignore.append(line.replace("Would remove ", "").strip())

        return ignore

    def _build_exclude_list(self, context):
        path = os.path.join(self.resource.name, ".rsync-exclude")

        ignore = [".rsync-exclude"]

        if os.path.isdir(self.resource.repository):
            svndir = os.path.join(self.resource.repository, ".svn")
            if os.path.isdir(svndir):
                ignore.append(".svn")
                ignore.extend(self._get_svn_ignore(context, self.resource.repository))

            gitdir = os.path.join(self.resource.repository, ".git")
            if os.path.isdir(gitdir):
                ignore.append(".git")
                ignore.extend(self._get_git_ignore(context, self.resource.repository))

            ignorefile = os.path.join(self.resource.repository, ".rsync-exclude")
            if os.path.exists(ignorefile):
                ignore.extend(x for x in open(ignorefile).read().split("\n") if x.strip())

        open(path, "w").write("\n".join(ignore))

        return path

    def _sync(self, context, dryrun=False):
        # FIXME: This will touch disk even in simulate mode... But *only* the exclude file.
        command = ["/usr/bin/rsync", "-rltv", "--stats", "--delete", "--exclude-from", self._build_exclude_list(context)]
        if dryrun:
            command.extend(["-n"])
        command.extend([".", self.resource.name+"/"])

        rv, out, err = context.shell.execute(command, cwd=self.resource.repository, user=self.resource.user, exceptions=False, passthru=dryrun)

        if context.simulate and not dryrun:
            # We won't get any output from _sync if we aren't doing a dry-run whilst in simulate mode
            # But this is only called with dryrun = False when there is stuff to sync
            return True

        if out.split("\n")[1].strip():
            return True

        return False

    def apply(self, context):
        changed = False

        if not os.path.exists("/usr/bin/rsync"):
            context.changelog.info("'/usr/bin/rsync' command is not available.")
            if not context.simulate:
                raise MissingDependency("Cannot continue with unmet dependency")

        if not os.path.exists(self.resource.name):
            command = ["/bin/mkdir", self.resource.name]
            rv, out, err = context.shell.execute(command, exceptions=False)
            changed = True

        if context.simulate and changed:
            context.changelog.info("Root directory not found; assuming rsync will occur")
            return True

        ac = AttributeChanger(context, self.resource.name, self.resource.user, mode=0755)
        ac.apply(context)
        changed = changed or ac.changed

        if self._sync(context, True):
            self._sync(context)
            changed = True

        return changed




# coding=utf-8

import unittest
from yaybu.providers.service.upstart import _UpstartServiceMixin

class TestUpstartParser(unittest.TestCase):

    """ Tests our ability to handle output from /sbin/status """

    def parse(self, block):
        u = _UpstartServiceMixin()
        return list(u._parse_status_output(block))

    def test_multiple(self):
        lines = [
            "network-interface-security (network-manager) start/running",
            "network-interface-security (network-interface/eth0) start/running",
            "network-interface-security (network-interface/lo) start/running",
            "network-interface-security (networking) start/running",
            ]
        output = self.parse("\n".join(lines))

        self.assertEqual(len(output), 4)
        self.assertEqual(output[1].name, "network-interface/eth0")
        self.assertEqual(output[2].goal, "start")
        self.assertEqual(output[3].status, "running")

    def test_with_instance_name(self):
        output = self.parse("network-interface-security (network-manager) start/running\n")

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0].name, "network-manager")
        self.assertEqual(output[0].goal, "start")
        self.assertEqual(output[0].status, "running")

    def test_start_running_with_pid(self):
        output = self.parse("ssh start/running, process 1234\n")

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0].name, "ssh")
        self.assertEqual(output[0].goal, "start")
        self.assertEqual(output[0].status, "running")

    def test_stop_waiting_no_pid(self):
        output = self.parse("hwclock stop/waiting\n")

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0].name, "hwclock")
        self.assertEqual(output[0].goal, "stop")
        self.assertEqual(output[0].status, "waiting")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import error
import datetime
import dateutil.parser
import types
import urlparse
import sys
import os
from abc import ABCMeta, abstractmethod, abstractproperty
import unicodedata
import random
import yay

unicode_glyphs = ''.join(
    unichr(char)
    for char in xrange(sys.maxunicode+1)
    if unicodedata.category(unichr(char))[0] in ('LMNPSZ')
    )


# we abuse urlparse for our parsing needs
urlparse.uses_netloc.append("package")

class Argument(object):

    """ Stores the argument value on the instance object. It's a bit fugly,
    neater ways of doing this that do not involve passing extra arguments to
    Argument are welcome. """

    metaclass = ABCMeta
    argument_id = 0

    def __init__(self, **kwargs):
        self.default = kwargs.pop("default", None)
        self.__doc__ = kwargs.pop("help", None)
        self.arg_id = "argument_%d" % Argument.argument_id
        Argument.argument_id += 1

    def __get__(self, instance, owner):
        if instance is None:
            # sphinx complains?
            #raise AttributeError
            return None
        if hasattr(instance, self.arg_id):
            return getattr(instance, self.arg_id)
        else:
            return self.default

    @abstractmethod
    def __set__(self, instance, value):
        """ Set the property. The value will be a UTF-8 encoded string read from the yaml source file. """

class Boolean(Argument):

    """ Represents a boolean. "1", "yes", "on" and "true" are all considered
    to be True boolean values. Anything else is False. """

    def __set__(self, instance, value):
        if type(value) in types.StringTypes:
            if value.lower() in ("1", "yes", "on", "true"):
                value = True
            else:
                value = False
        else:
            value = bool(value)
        setattr(instance, self.arg_id, value)

class String(Argument):

    """ Represents a string. """

    def __set__(self, instance, value):
        if value is None:
            pass
        elif not isinstance(value, (unicode, yay.String)):
            value = unicode(value, 'utf-8')
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        l = []
        for i in range(random.randint(0, 1024)):
            l.append(random.choice(unicode_glyphs))
        return "".join(l)

class FullPath(Argument):

    """ Represents a full path on the filesystem. This should start with a
    '/'. """

    def __set__(self, instance, value):
        if value is None:
            pass
        elif not isinstance(value, unicode):
            value = unicode(value, 'utf-8')
        if not value.startswith("/"):
            raise error.ParseError("%s is not a full path" % value)
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        # TODO: needs work
        l = []
        for i in range(random.randint(0, 1024)):
            l.append(random.choice(unicode_glyphs))
        return "/" + "".join(l)

class Integer(Argument):

    """ Represents an integer argument taken from the source file. This can
    throw an :py:exc:error.ParseError if the passed in value cannot represent
    a base-10 integer. """

    def __set__(self, instance, value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except ValueError:
                raise error.ParseError("%s is not an integer" % value)
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        return random.randint(0,sys.maxint)

class DateTime(Argument):

    """ Represents a date and time. This is parsed in ISO8601 format. """

    def __set__(self, instance, value):
        if isinstance(value, basestring):
            value = dateutil.parser.parse(value)
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        return datetime.datetime.fromtimestamp(random.randint(0, sys.maxint))

class Octal(Integer):

    """ An octal integer.  This is specifically used for file permission modes. """

    def __set__(self, instance, value):
        if isinstance(value, int):
            # we assume this is due to lame magic in yaml and rebase it
            value = int(str(value), 8)
        else:
            value = int(value, 8)
        setattr(instance, self.arg_id, value)

class Dict(Argument):
    def __set__(self, instance, value):
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        return {}

class List(Argument):
    def __set__(self, instance, value):
        setattr(instance, self.arg_id, value)

    @classmethod
    def _generate_valid(self):
        return []


class File(Argument):

    """ Provided with a URL, this can get files by various means. Often used
    with the package:// scheme """

    def __set__(self, instance, value):
        setattr(instance, self.arg_id, value)


class StandardPolicy:

    def __init__(self, policy_name):
        self.policy_name = policy_name

class PolicyTrigger:

    def __init__(self, policy, when, on, immediately=True):
        self.policy = policy
        self.when = when
        self.on = on
        self.immediately = immediately

    def bind(self, resources, target):
        if self.on in resources:
            resources[self.on].register_observer(self.when, target, self.policy, self.immediately)
        else:
            raise error.BindingError("Cannot bind %r to missing resource named '%s'" % (target, self.on))
        return resources[self.on]

class PolicyCollection:

    """ A collection of policy structures. """

    literal = None
    """ The policy that is set as the "standard" policy, not one that depends on a trigger. """

    triggers = ()
    """ A list of PolicyTrigger objects that represent optional triggered policies. """

    def __init__(self, literal=None, triggers=()):
        self.literal = literal
        self.triggers = triggers

    def literal_policy(self, resource):
        if self.literal is not None:
            return resource.policies[self.literal.policy_name]
        else:
            import policy
            return policy.NullPolicy

class PolicyArgument(Argument):

    """ Parses the policy: argument for resources, including triggers etc. """

    def __set__(self, instance, value):
        """ Set either a default policy or a set of triggers on the policy collection """
        if type(value) in types.StringTypes:
            coll = PolicyCollection(StandardPolicy(value))
        else:
            triggers = []
            for policy, conditions in value.items():
                if not isinstance(conditions, list):
                    conditions = [conditions]
                for condition in conditions:
                    triggers.append(
                        PolicyTrigger(
                            policy=policy,
                            when=condition['when'],
                            on=condition['on'],
                            immediately=condition.get('immediately', 'true') == 'true')
                        )
            coll = PolicyCollection(triggers=triggers)
        setattr(instance, self.arg_id, coll)



from yaybu.harness import FakeChrootTestCase
from time import sleep

class TestPackageInstallation(FakeChrootTestCase):

    def test_already_installed(self):
        rv = self.fixture.apply("""
            resources:
              - Package:
                  name: python
            """)
        self.assertEqual(rv, 254)

    def test_installation(self):
        self.fixture.check_apply("""
            resources:
              - Package:
                  name: hello
            """)

    def test_nonexistent_package(self):
        """ Try to install a package that does not exist. """
        rv = self.fixture.apply("""
            resources:
              - Package:
                  name: zzzz
            """)
        self.assertEqual(rv, 132)

    def test_package_reinstallation(self):
        """ Try reinstalling a previously-removed package """
        hello_install = """
            resources:
              - Package:
                  name: hello
            """

        hello_remove = """
            resources:
              - Package:
                  name: hello
                  policy: uninstall
            """

        self.fixture.check_apply(hello_install)
        self.fixture.check_apply(hello_remove)
        self.fixture.check_apply(hello_install)


class TestPackageRemoval(FakeChrootTestCase):

    def test_installed(self):
        """ Try removing a package that is installed. """
        self.fixture.check_apply("""
            resources:
              - Package:
                  name: ubuntu-keyring
                  policy: uninstall
            """)





from yaybu.harness import FakeChrootTestCase
from yaybu.core.error import MissingDependency

class RsyncTest(FakeChrootTestCase):
    """
    Test the rsync checkout provider.

    To run these tests, rsync must be installed in the test environment.
    """

    def setUp(self):
        super(RsyncTest, self).setUp()
        self.sync()

    def sync(self, expect=0):
        return self.fixture.check_apply("""
            resources:
                - Package:
                    name: rsync
                - Directory:
                  - name: /source
                    mode: 0755
                - Checkout:
                    scm: rsync
                    name: /dest
                    repository: /source
            """, expect=expect)


    def test_add(self):
        self.fixture.touch("/source/a")
        self.sync()
        self.failUnless(self.fixture.exists("/dest/a"))

    def test_del(self):
        self.fixture.touch("/dest/b")
        self.sync()
        self.failUnless(not self.fixture.exists("/dest/b"))

    def test_nochange(self):
        self.sync(expect=254)


class RsyncMissingTest(FakeChrootTestCase):

    def test_missing_rsync(self):
        rv = self.fixture.apply("""
           resources:
               - Checkout:
                   scm: rsync
                   name: /dest
                   repository: /source
           """)
        self.assertEqual(MissingDependency.returncode, rv)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from fixtures import Fixture as BaseFixture
except ImportError:
    class BaseFixture(object):
        """
        I am a Fixture compatible with the fixtures API by Robert Collins
        If the fixtures package is installed I won't be used.
        """
        def getDetails(self):
            return {}


class Fixture(BaseFixture):

    """ A base class for Fixtures that providing virtual environments to deploy configuration in.

    This base class is abstract and provides no concrete implementations. For a simple implementation
    of this interface see :py:class:`~yaybu.harness.fakechroot.FakeChrootFixture`.
    """

    def exists(self, path):
        """ Checks whether or not a path exists in the target """
        raise NotImplementedError(self.exists)

    def isdir(self, path):
        """ Checks whether or not a path is a directory in the target """
        raise NotImplementedError(self.isdir)

    def mkdir(self, path):
        """ Creates a directory in the target """
        raise NotImplementedError(self.mkdir)

    def open(self, path, mode='r'):
        """ Opens a file in the target """
        raise NotImplementedError(self.open)

    def touch(self, path):
        """ Ensures that a file exists in the target """
        raise NotImplementedError(self.touch)

    def chmod(self, path, mode):
        """ Change the permissions of a path in the target """
        raise NotImplementedError(self.chmod)

    def readlink(self, path):
        """ Return a string containing the path that a symbolic link points to """
        raise NotImplementedError(self.readlink)

    def symlink(self, source, dest):
        """ Create a symbolic link pointing to source at dest """
        raise NotImplementedError(self.symlink)

    def stat(self, path):
        """ Perform the equivalent of the a stat() system call on the given path """
        raise NotImplementedError(self.stat)




from yaybu.harness import FakeChrootTestCase
from yaybu.core import error
import pwd
import grp
import os
import stat
import errno

def sibpath(filename):
    return os.path.join(os.path.dirname(__file__), filename)

class TestFileApply(FakeChrootTestCase):

    def test_create_missing_component(self):
        rv = self.fixture.apply("""
            resources:
              - File:
                  name: /etc/missing/filename
            """)
        self.assertEqual(rv, error.PathComponentMissing.returncode)

    def test_create_missing_component_simulate(self):
        """
        Right now we treat missing directories as a warning in simulate mode, as other outside processes might have created them.
        Later on we might not generate warnings for resources we can see will be created
        """
        rv = self.fixture.apply_simulate("""
            resources:
              - File:
                  name: /etc/missing/filename
            """)
        self.assertEqual(rv, 0)

    def test_create_file(self):
        self.fixture.check_apply("""
            resources:
              - File:
                  name: /etc/somefile
                  owner: root
                  group: root
            """)

        self.failUnlessExists("/etc/somefile")

    def test_attributes(self):
        self.fixture.check_apply("""
            resources:
              - File:
                  name: /etc/somefile2
                  owner: nobody
                  group: nogroup
                  mode: 0666
            """)
        self.failUnlessExists("/etc/somefile2")
        st = self.fixture.stat("/etc/somefile2")
        self.failUnless(pwd.getpwuid(st.st_uid)[0] != 'nobody')
        self.failUnless(grp.getgrgid(st.st_gid)[0] != 'nogroup')
        mode = stat.S_IMODE(st.st_mode)
        self.assertEqual(mode, 0666)

    def test_create_file_template(self):
        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/templated
                    template: package://yaybu.providers.tests/template1.j2
                    template_args:
                        foo: this is foo
                        bar: 42
                    owner: root
                    group: root
                    """)
        self.failUnlessExists("/etc/templated")

    def test_create_file_template_with_extends(self):
        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/templated
                    template: package://yaybu.providers.tests/template_with_extends.j2
                    template_args:
                        foo: this is foo
                        bar: 42
                    owner: root
                    group: root
                    """)
        self.failUnlessExists("/etc/templated")
        with self.fixture.open("/etc/templated") as fp:
            self.failUnless("this is foo" in fp.read())

    def test_modify_file(self):
        with self.fixture.open("/etc/test_modify_file", "w") as fp:
          fp.write("foo\nbar\nbaz")

        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/test_modify_file
                    template: package://yaybu.providers.tests/template1.j2
                    template_args:
                        foo: this is a modified file
                        bar: 37
            """)

    def test_remove_file(self):
        self.fixture.check_apply("""
            resources:
              - File:
                  name: /etc/toremove
            """)
        self.fixture.check_apply("""
            resources:
              - File:
                  name: /etc/toremove
                  policy: remove
            """)
        self.failIfExists("/etc/toremove")


    def test_empty(self):
        with self.fixture.open("/etc/foo", "w") as fp:
            fp.write("foo")

        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/foo
            """)

    def test_empty_nochange(self):
        with self.fixture.open("/etc/foo", "w") as fp:
            fp.write("")

        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/foo
            """)
        self.assertEqual(rv, 254)


    def test_carriage_returns(self):
        """ a template that does not end in \n will still result in a file ending in \n """
        with self.fixture.open("/etc/test_carriage_returns", "w") as fp:
            fp.write("foo\n")

        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/test_carriage_returns
                    template: package://yaybu.providers.tests/test_carriage_returns.j2
            """)
        self.assertEqual(rv, 254) # nothing changed

    def test_carriage_returns2(self):
        """ a template that does end in \n will not gain an extra \n in the resulting file"""
        with self.fixture.open("/etc/test_carriage_returns2", "w") as fp:
            fp.write("foo\n")

        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/test_carriage_returns2
                    template: package://yaybu.providers.tests/test_carriage_returns2.j2
            """)
        self.assertEqual(rv, 254) # nothing changed

    def test_unicode(self):
        self.fixture.check_apply(open(sibpath("unicode1.yay")).read())

    def test_static(self):
        """ Test setting the contents to that of a static file. """
        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/foo
                    static: package://yaybu.providers.tests/test_carriage_returns2.j2
            """)

    def test_missing(self):
        """ Test trying to use a file that isn't in the yaybu path """
        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/foo
                    static: this-doesnt-exist
            """)
        self.failUnlessEqual(rv, error.MissingAsset.returncode)


class TestFileRemove(FakeChrootTestCase):

    def test_remove(self):
        """ Test removing a file that exists. """
        with self.fixture.open("/etc/bar","w") as fp:
            fp.write("")

        self.fixture.check_apply("""
            resources:
                - File:
                    name: /etc/bar
                    policy: remove
            """)

    def test_remove_missing(self):
        """ Test removing a file that does not exist. """
        self.failIfExists("/etc/baz")
        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/baz
                    policy: remove
            """)
        self.failUnlessEqual(rv, 254)

    def test_remove_notafile(self):
        """ Test removing something that is not a file. """
        self.fixture.mkdir("/etc/qux")
        rv = self.fixture.apply("""
            resources:
                - File:
                    name: /etc/qux
                    policy: remove
            """)
        self.failUnlessEqual(rv, 139)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os, logging

from yaybu.core.shell import Command
from yaybu.core.provider import Provider
from yaybu.core.error import MissingDependency
from yaybu import resources

import shlex


log = logging.getLogger("subversion")

class Svn(Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    @classmethod
    def isvalid(self, policy, resource, yay):
        identities = [
            'svn',
            'subversion',
        ]

        if resource.scm:
            return resource.scm.lower() in identities

        # TODO: ensure that this should be made default provider
        return True

    @property
    def url(self):
        return self.resource.repository + "/" + self.resource.branch

    def action_checkout(self, context):
        if os.path.exists(self.resource.name):
            return

        log.debug("Checking out %s" % self.resource)
        self.svn(context, "co", self.url, self.resource.name)
        return True

    def apply(self, context):
        if not os.path.exists("/usr/bin/svn"):
            error_string = "'/usr/bin/svn' is not available; update your configuration to install subversion?"
            if not context.simulate:
                raise MissingDependency(error_string)
            log.info(error_string)
            log.info("This error was ignored in simulate mode")

        if not os.path.exists(self.resource.name):
            return self.action_checkout(context)

        log.debug("Syncing %s" % self.resource)

        changed = False

        info = self.info(context, self.resource.name)
        repo_info = self.info(context, self.url)

        # If the 'Repository Root' is different between the checkout and the repo, switch --relocated
        old_repo_root = info["Repository Root"]
        new_repo_root = repo_info["Repository Root"]
        if old_repo_root != new_repo_root:
            log.info("Switching repository root from '%s' to '%s'" % (old_repo_root, new_repo_root))
            self.svn(context, "switch", "--relocate", old_repo_root, new_repo_root, self.resource.name)
            changed = True

        # If we have changed branch, switch
        old_url = info["URL"]
        new_url = repo_info["URL"]
        if old_url != new_url:
            log.info("Switching branch from '%s' to '%s'" % (old_url, new_url))
            self.svn(context, "switch", new_url, self.resource.name)
            changed = True

        # If we have changed revision, svn up
        # FIXME: Eventually we might want revision to be specified in the resource?
        current_rev = info["Last Changed Rev"]
        target_rev = repo_info["Last Changed Rev"]
        if current_rev != target_rev:
            log.info("Switching revision from %s to %s" % (current_rev, target_rev))
            self.svn(context, "up", "-r", target_rev, self.resource.name)
            changed = True

        return changed

    def action_export(self, context):
        if os.path.exists(self.resource.name):
            return
        log.debug("Exporting %s" % self.resource)
        self.svn(context, "export", self.url, self.resource.name)

    def get_svn_args(self, action, *args, **kwargs):
        command = Command(["svn"])

        if kwargs.get("quiet", False):
            command.add("--quiet")

        command.extend([action, "--non-interactive"])

        if self.resource.scm_username:
            command.add(Command(["--username", self.resource.scm_username]))
        if self.resource.scm_password:
            command.add(Command(["--password", self.resource.scm_password]))
        if self.resource.scm_username or self.resource.scm_password:
            command.add("--no-auth-cache")

        for arg in args:
            command.add(arg)

        return command

    def info(self, context, uri):
        command = self.get_svn_args("info", uri)
        returncode, stdout, stderr = context.shell.execute(command, passthru=True)
        return dict(x.split(": ") for x in stdout.split("\n") if x)

    def svn(self, context, action, *args, **kwargs):
        command = self.get_svn_args(action, *args, **kwargs)
        return context.shell.execute(command, user=self.resource.user)





# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from yaybu.core import error
from yaybu import resources

# http://upstart.ubuntu.com/getting-started.html
# http://upstart.ubuntu.com/cookbook/#initctl-status

class _ServiceMixin(object):

    features = []

    def status(self, context):
        if self.resource.running:
            returncode, stdout, stderr = context.shell.execute(self.resource.running, passthru=True, exceptions=False)
            if returncode != 0:
                return "not-running"
            else:
                return "running"

        if not self.resource.pidfile:
            return "unknown"

        if not os.path.exists(self.resource.pidfile):
            return "not-running"

        pid = open(self.resource.pidfile).read().strip()
        try:
            pid = int(pid)
        except:
            return "unknown"

        try:
            os.kill(pid, 0)
            return "running"
        except OSError, e:
            return "not-running"

    def do(self, context, action):
        returncode, stdout, stderr = context.shell.execute(self.get_command(action), exceptions=False)

        if returncode != 0:
            raise error.CommandError("%s failed with return code %d" % (action, returncode))

    def ensure_enabled(self, context):
        pass

    def ensure_disabled(self, context):
        pass


class _Start(object):

    policies = (resources.service.ServiceStartPolicy,)

    def apply(self, context):
        changed = self.ensure_enabled(context)

        if self.status(context) == "running":
            return changed

        self.do(context, "start")

        return True


class _Stop(object):

    policies = (resources.service.ServiceStopPolicy,)

    def apply(self, context):
        changed = self.ensure_disabled(context)

        if self.status(context) == "not-running":
            return changed

        self.do(context, "stop")

        return True


class _Restart(object):

    policies = (resources.service.ServiceRestartPolicy,)

    def apply(self, context):
        changed = self.ensure_enabled(context)

        if self.status(context) == "not-running":
            self.do(context, "start")
            return True

        if "restart" in self.features:
            self.do(context, "restart")
        else:
            self.do(context, "stop")
            self.do(context, "start")

        return True






""" Classes that handle logging of changes. """

import abc
import sys
import logging
import types
import json

from yaybu.core import error

logger = logging.getLogger("audit")


class ResourceFormatter(logging.Formatter):

    """ Automatically add a header and footer to log messages about particular
    resources """

    def __init__(self, *args):
        logging.Formatter.__init__(self, *args)
        self.resource = None

    def format(self, record):
        next_resource = getattr(record, "resource", None)

        rv = u""

        # Is the logging now about a different resource?
        if self.resource != next_resource:

            # If there was already a resource, let us add a footer
            if self.resource:
                rv += self.render_resource_footer()

            self.resource = next_resource

            # Are we now logging for a new resource?
            if self.resource:
                rv += self.render_resource_header()

        formatted = logging.Formatter.format(self, record)

        if self.resource:
            rv += "\r\n".join("| %s" % line for line in formatted.splitlines()) + "\r"
        else:
            rv += formatted

        return rv

    def render_resource_header(self):
        header = unicode(self.resource)

        rl = len(header)
        if rl < 80:
            total_minuses = 77 - rl
            minuses = total_minuses/2
            leftover = total_minuses % 2
        else:
            minuses = 4
            leftover = 0

        return u"/%s %s %s\n" % ("-"*minuses,
                                 header,
                                 "-"*(minuses + leftover))

    def render_resource_footer(self):
        return u"\%s\n\n" % ("-" *79,)


class Change(object):
    """ Base class for changes """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def apply(self, renderer):
        """ Apply the specified change. The supplied renderer will be
        instantiated as below. """

class AttributeChange(Change):
    """ A change to one attribute of a file's metadata """

class ChangeRendererType(type):

    """ Keeps a registry of available renderers by type. The only types
    supported are text """

    renderers = {}

    def __new__(meta, class_name, bases, new_attrs):
        cls = type.__new__(meta, class_name, bases, new_attrs)
        if cls.renderer_for is not None:
            ChangeRendererType.renderers[(cls.renderer_type, cls.renderer_for)] = cls
        return cls

class ChangeRenderer:

    """ A class that knows how to render a change. """

    __metaclass__ = ChangeRendererType

    renderer_for = None
    renderer_type = None

    def __init__(self, logger, verbose):
        self.logger = logger
        self.verbose = verbose

    def render(self, logger):
        pass

class TextRenderer(ChangeRenderer):
    renderer_type = "text"

class ResourceChange(object):

    """ A context manager that handles logging per resource. This allows us to
    elide unchanged resources, which is the default logging output option. """

    def __init__(self, changelog, resource):
        self.changelog = changelog
        self.resource = resource

        # We wrap the logger so it always has context information
        logger = logging.getLogger("yaybu.changelog")
        self.logger = logging.LoggerAdapter(logger, dict(resource=unicode(resource)))

    def info(self, message, *args):
        self.logger.info(message, *args)

    def notice(self, message, *args):
        self.logger.info(message, *args)

    def __enter__(self):
        self.changelog.current_resource = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exc_type = exc_type
        self.exc_val = exc_val
        self.exc_tb = exc_tb

        if self.exc_val is not None:
            self.notice("Exception: %s" % (self.exc_val,))

        self.changelog.current_resource = None


class ChangeLog:

    """ Orchestrate writing output to a changelog. """

    def __init__(self, context):
        self.current_resource = None
        self.ctx = context
        self.verbose = self.ctx.verbose

        self.logger = logging.getLogger("yaybu.changelog")

        self.configure_session_logging()
        self.configure_audit_logging()

    def configure_session_logging(self):
        root = logging.getLogger("yaybu.changelog")
        root.setLevel(logging.INFO)

        if len(root.handlers):
            # Session logging has already been configured elsewhere?
            return

        handler = logging.StreamHandler(sys.stdout)
        #handler.setFormatter(logging.Formatter("%(message)s"))
        handler.setFormatter(ResourceFormatter("%(message)s"))
        root.addHandler(handler)

    def configure_audit_logging(self):
        """ configure the audit trail to log to file or to syslog """

        if self.ctx.simulate:
            return

        options = self.ctx.options.get("auditlog", {})
        mode = options.get("mode", "off")

        if mode == "off":
            return

        levels = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
            }

        level = levels.get(options.get("level", "info"), None)
        if level is None:
            raise KeyError("Log level %s not recognised, terminating" % option["level"])

        root = logging.getLogger()

        if mode == "file":
            handler = logging.FileHandler(options.get("logfile", "/var/log/yaybu.log"))
            #handler.setFormatter(logging.Formatter("%(message)s"))
            handler.setFormatter(ResourceFormatter("%(asctime)s %(message)s"))
            root.addHandler(handler)

        elif mode == "syslog":
            facility = getattr(logging.handlers.SysLogHandler, "LOG_LOCAL%s" % options.get("facility", "7"))
            handler = logging.handlers.SysLogHandler("/dev/log", facility=facility)
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            handler.setFormatter(formatter)
            logging.getLogger().addHandler(handler)

    def write(self, line=""):
        #FIXME: Very much needs removing
        self.logger.info(line)

    def resource(self, resource):
        return ResourceChange(self, resource)

    def apply(self, change):
        """ Execute the change, passing it the appropriate renderer to use. """
        renderers = []
        text_class = ChangeRendererType.renderers.get(("text", change.__class__), None)
        return change.apply(text_class(self, self.verbose))

    def info(self, message, *args, **kwargs):
        """ Write a textual information message. This is used for both the
        audit trail and the text console log. """
        if self.current_resource:
            self.current_resource.info(message, *args)
        else:
            self.logger.info(message, *args)

    def notice(self, message, *args, **kwargs):
        """ Write a textual notification message. This is used for both the
        audit trail and the text console log. """
        if self.current_resource:
            self.current_resource.notice(message, *args)
        else:
            self.logger.info(message, *args)

    def debug(self, message, *args, **kwargs):
        pass

    def error(self, message, *args):
        self.logger.error(message, *args)

    def handle(self, record):
        self.logger.handle(record)


class RemoteHandler(logging.Handler):

    def __init__(self, connection):
        logging.Handler.__init__(self)
        self.connection = connection

    def emit(self, record):
        data = json.dumps(record.__dict__)

        self.connection.request("POST", "/changelog/", data, {"Content-Length": len(data)})
        rsp = self.connection.getresponse()

        lngth = rsp.getheader("Content-Length", 0)
        rsp.read(lngth)


class RemoteChangeLog(ChangeLog):

    def configure_session_logging(self):
        root = logging.getLogger()
        root.setLevel(logging.INFO)

        handler = RemoteHandler(self.ctx.connection)
        handler.setFormatter(logging.Formatter("%(message)s"))
        root.addHandler(handler)

    def configure_audit_logging(self):
        # We don't want to try and log changes in syslog on the box we are pressing go on,
        # only the box we are deploying to. So no audit logging here.
        pass




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex

from yaybu.core import provider
from yaybu.providers.service import utils


class _SimpleServiceMixin(utils._ServiceMixin):

    features = ["restart", ]

    @classmethod
    def isvalid(cls, policy, resource, yay):
        if not super(_SimpleServiceMixin, cls).isvalid(policy, resource, yay):
            return False
        if not getattr(resource, policy.name):
            return False
        return True

    def get_command(self, action):
        return shlex.split(getattr(self.resource, action).encode("UTF-8"))


class Start(_SimpleServiceMixin, utils._Start, provider.Provider):
    pass

class Stop(_SimpleServiceMixin, utils._Stop, provider.Provider):
    pass

class Restart(_SimpleServiceMixin, utils._Restart, provider.Provider):
    pass




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import error

class PolicyType(type):

    """ Registers the policy on the resource """

    def __new__(meta, class_name, bases, new_attrs):
        cls = type.__new__(meta, class_name, bases, new_attrs)
        cls.providers = []
        if cls.resource is not None:
            cls.resource.policies[cls.name] = cls
        return cls

class Policy(object):

    """
    A policy is a representation of a resource. A policy requires a
    certain argument signature to be present before it can be used. There may
    be multiple policies selected for a resource, in which case all argument
    signatures must be conformant.

    Providers must provide all selected policies to be a valid provider for
    the resource.
    """

    __metaclass__ = PolicyType

    # specify true if you wish this policy to be enabled by default
    default = False
    # the name of the policy, used to find it in the ensures config
    name = None
    # specify the resource to which this policy applies
    resource = None
    # the list of providers that provide this policy
    providers = []
    # Override this with a list of assertions
    signature = ()

    def __init__(self, resource):
        self.resource = resource

    @classmethod
    def conforms(self, resource):
        """ Test if the provided resource conforms to the signature for this
        policy. """
        for a in self.signature:
            if not a.test(resource):
                return False
        return True

    def get_provider(self, yay):
        """ Get the one and only one provider that is valid for this resource,
        policy and overall context """
        valid = [p.isvalid(self, self.resource, yay) for p in self.providers]
        if valid.count(True) > 1:
            raise error.TooManyProviders()
        if valid.count(True) == 0:
            raise error.NoSuitableProviders()
        return self.providers[valid.index(True)]

class NullPolicy(Policy):
    pass

class ArgumentAssertion(object):

    """ An assertion of the state of an argument """

    def __init__(self, name):
        self.name = name

class Present(ArgumentAssertion):

    """ The argument has been specified, or has a default value. """

    def test(self, resource):
        """ Test that the argument this asserts for is present in the
        resource. """
        if getattr(resource, self.name) is not None:
            return True
        return False

class Absent(ArgumentAssertion):

    """ The argument has not been specified by the user and has no default
    value. An argument with a default value is always defined. """

    def test(self, resource):
        if getattr(resource, self.name) is None:
            return True
        return False

class AND(ArgumentAssertion):

    def __init__(self, *args):
        self.args = args

    def test(self, resource):
        for a in self.args:
            if not a.test(resource):
                return False
        return True

class NAND(ArgumentAssertion):

    def __init__(self, *args):
        self.args = args

    def test(self, resource):
        results = [1 for a in self.args if a.test(resource)]
        if len(results) > 1:
            return False
        return True

class XOR(ArgumentAssertion):

    def __init__(self, *args):
        self.args = args

    def test(self, resource):
        l = [1 for a in self.args if a.test(resource)]
        if len(l) == 0:
            return False
        elif len(l) == 1:
            return True
        else:
            return False



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import testtools


class TestCase(testtools.TestCase):

    def useFixture(self, fixture):
        """ Use a fixture in a test case.

        The fixture will be setUp, and self.addCleanup(fixture.cleanUp) called.

        The fixture will be available as self.fixture.

        :param fixture: The fixture to use
        :return: The fixture, after setting it up and scheduling a cleanup for it
        """
        self.fixture = fixture
        return super(TestCase, self).useFixture(fixture)

    def failUnlessExists(self, path):
        if not self.fixture.exists(path):
            self.fail("Path '%s' does not exist" % path)

    def failIfExists(self, path):
        if self.fixture.exists(path):
            self.fail("Path '%s' exists" % path)




import os, shutil, grp, signal

from yaybu.harness import FakeChrootTestCase
from yaybu.util import sibpath

simpleservice = """
#! /usr/bin/env python
import os, select, sys
if __name__ == "__main__":
    if os.fork() != 0:
        os._exit(0)

    os.setsid()

    if os.fork() != 0:
        os._exit(0)

    open("simple_daemon.pid", "w").write(str(os.getpid()))
    #os.chdir("/")
    os.umask(0)

    for fd in range(0, 1024):
        try:
            os.close(fd)
        except OSError:
            pass

    os.open("/dev/null", os.O_RDWR)

    os.dup2(0, 1)
    os.dup2(0, 2)

    while True:
        select.select([sys.stdin], [], [])
"""


class TestSimpleService(FakeChrootTestCase):

    def setUp(self):
        super(TestSimpleService, self).setUp()

        with self.fixture.open("/bin/simple_daemon", "w") as fp:
            fp.write(simpleservice)

    def test_start(self):
        self.fixture.check_apply("""
            resources:
                - Service:
                    name: test
                    policy: start
                    start: python /bin/simple_daemon
                    pidfile: /simple_daemon.pid
            """)

        with self.fixture.open("/simple_daemon.pid") as fp:
            pid = int(fp.read())

        os.kill(pid, signal.SIGTERM)

    def test_stop(self):
        self.fixture.call(["python", "/bin/simple_daemon"])

        self.fixture.check_apply("""
            resources:
                - Service:
                    name: test
                    policy: stop
                    stop: sh -c 'kill $(cat /simple_daemon.pid)'
                    pidfile: /simple_daemon.pid
            """)

    def test_restart(self):
        rv = self.fixture.apply("""
            resources:
                - Service:
                    name: test
                    policy: restart
                    restart: touch /foo
            """)

        # We restart every time config is applied - so check_apply would fail the
        # automatic idempotentcy check
        self.failUnlessEqual(rv, 0)
        self.failUnlessExists("/foo")

    def test_running_true(self):
        rv = self.fixture.apply("""
            resources:
                - Service:
                    name: test
                    start: touch /test_running_true
                    running: /bin/sh -c "true"
            """)

        self.failUnlessEqual(rv, 254)

    def test_running_false(self):
        rv = self.fixture.apply("""
            resources:
                - Service:
                    name: test
                    start: touch /test_running_false
                    running: /bin/sh -c "false"
            """)

        self.failUnlessEqual(rv, 0)
        self.failUnlessExists("/test_running_false")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaybu.core.resource import Resource
from yaybu.core.policy import (
    Policy,
    Absent,
    Present,
    )
from yaybu.core.argument import (
    String,
    Integer,
    )

class Prompt(Resource):

    """ Ask a question of the operator. This is useful if manual steps are
    required as part of implementing a configuration. For example, changes to
    off-system web interfaces or databases may be required as part of applying
    a change. """

    name = String()
    """ A unique descriptive name for the resource. """

    question = String()
    """ The question to ask the operator. """

class PromptPolicy(Policy):

    """ Prompt the operator.

    The value of the question attribute will be displayed to the operator and
    deployment will not continue until they acknowledge the prompt."""

    resource = Prompt
    name = "prompt"
    default = True
    signature = (
        Present("name"),
        Present("question"),
        )



import os
import sys
from yaybu.harness import FakeChrootTestCase
from yaybu.core import error


class TestWatched(FakeChrootTestCase):

    def test_watched(self):
        self.fixture.check_apply("""
            resources:
                - Execute:
                    name: test_watched
                    command: touch /watched-file
                    creates: /watched-file
                    watch:
                      - /watched-file
                - Execute:
                    name: test_output
                    command: touch /event-triggered
                    creates: /event-triggered
                    policy:
                        execute:
                            when: watched
                            on: File[/watched-file]
            """)
        self.failUnlessExists("/event-triggered")




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import stat
import pwd
import grp
import difflib
import logging
import string

try:
    import magic
except ImportError:
    magic = None

from jinja2 import Environment, BaseLoader, TemplateNotFound

from yaybu import resources
from yaybu.core import provider
from yaybu.core import change
from yaybu.core import error

from yay import String


def binary_buffers(*buffers):

    """ Check all of the passed buffers to see if any of them are binary. If
    any of them are binary this will return True. """
    if not magic:
        check = lambda buff: len(buff) == sum(1 for c in buff if c in string.printable)
    else:
        ms = magic.open(magic.MAGIC_MIME)
        ms.load()
        check = lambda buff: ms.buffer(buff).startswith("text/")

    for buff in buffers:
        if buff and not check(buff):
            return True
    return False

class AttributeChanger(change.Change):

    """ Make the changes required to a file's attributes """

    def __init__(self, context, filename, user=None, group=None, mode=None):
        self.context = context
        self.filename = filename
        self.user = user
        self.group = group
        self.mode = mode
        self.changed = False

    def apply(self, renderer):
        """ Apply the changes """
        exists = False
        uid = None
        gid = None
        mode = None

        if os.path.exists(self.filename):
            exists = True
            st = os.stat(self.filename)
            uid = st.st_uid
            gid = st.st_gid
            mode = stat.S_IMODE(st.st_mode)

        if self.user is not None:
            try:
                owner = pwd.getpwnam(self.user)
            except KeyError:
                if not self.context.simulate:
                    raise error.InvalidUser("User '%s' not found" % self.user)
                self.context.changelog.info("User '%s' not found; assuming this recipe will create it" % self.user)
                owner = None

            if not owner or owner.pw_uid != uid:
                self.context.shell.execute(["/bin/chown", self.user, self.filename])
                self.changed = True

        if self.group is not None:
            try:
                group = grp.getgrnam(self.group)
            except KeyError:
                if not self.context.simulate:
                    raise error.InvalidGroup("No such group '%s'" % self.group)
                self.context.changelog.info("Group '%s' not found; assuming this recipe will create it" % self.group) #FIXME
                group = None

            if not group or group.gr_gid != gid:
                self.context.shell.execute(["/bin/chgrp", self.group, self.filename])
                self.changed = True

        if self.mode is not None and mode is not None:
            if mode != self.mode:
                self.context.shell.execute(["/bin/chmod", "%o" % self.mode, self.filename])

                # Clear the user and group bits
                # We don't need to set them as chmod will *set* this bits with an octal
                # but won't clear them without a symbolic mode
                if mode & stat.S_ISGID and not self.mode & stat.S_ISGID:
                    self.context.shell.execute(["/bin/chmod", "g-s", self.filename])
                if mode & stat.S_ISUID and not self.mode & stat.S_ISUID:
                    self.context.shell.execute(["/bin/chmod", "u-s", self.filename])

                self.changed = True


class AttributeChangeRenderer(change.TextRenderer):
    renderer_for = AttributeChanger


class FileContentChanger(change.Change):

    """ Apply a content change to a file in a managed way. Simulation mode is
    catered for. Additionally the minimum changes required to the contents are
    applied, and logs of the changes made are recorded. """

    def __init__(self, context, filename, contents, sensitive):
        self.context = context
        self.filename = filename
        self.current = ""
        self.contents = contents
        self.changed = False
        self.renderer = None
        self.sensitive = sensitive

    def empty_file(self):
        """ Write an empty file """
        exists = os.path.exists(self.filename)
        if not exists:
            self.context.shell.execute(["touch", self.filename])
            self.changed = True
        else:
            st = os.stat(self.filename)
            if st.st_size != 0:
                self.renderer.empty_file(self.filename)
                if not self.context.simulate:
                    open(self.filename, "w").close()
                self.changed = True

    def overwrite_existing_file(self):
        """ Change the content of an existing file """
        self.current = open(self.filename).read()
        if self.current != self.contents:
            self.renderer.changed_file(self.filename, self.current, self.contents, self.sensitive)
            if not self.context.simulate:
                open(self.filename, "w").write(self.contents)
            self.changed = True

    def write_new_file(self):
        """ Write contents to a new file. """
        self.renderer.new_file(self.filename, self.contents, self.sensitive)
        if not self.context.simulate:
            open(self.filename, "w").write(self.contents)
        self.changed = True

    def write_file(self):
        """ Write to either an existing or new file """
        exists = os.path.exists(self.filename)
        if exists:
            self.overwrite_existing_file()
        else:
            self.write_new_file()

    def apply(self, renderer):
        """ Apply the changes necessary to the file contents. """
        self.renderer = renderer
        if self.contents is None:
            self.empty_file()
        else:
            self.write_file()


class FileChangeTextRenderer(change.TextRenderer):
    renderer_for = FileContentChanger

    def empty_file(self, filename):
        self.logger.notice("Emptied file %s", filename)

    def new_file(self, filename, contents, sensitive):
        self.logger.notice("Writting new file '%s'" % filename)
        if not sensitive:
            self.diff("", contents)

    def changed_file(self, filename, previous, replacement, sensitive):
        self.logger.notice("Changed file %s", filename)
        if not sensitive:
            self.diff(previous, replacement)

    def diff(self, previous, replacement):
        if not binary_buffers(previous, replacement):
            diff = "".join(difflib.unified_diff(previous.splitlines(1), replacement.splitlines(1)))
            for l in diff.splitlines():
                self.logger.info("    %s", l)
        else:
            self.logger.notice("Binary contents; not showing delta")


class YaybuTemplateLoader(BaseLoader):

    def __init__(self, ctx):
        self.ctx = ctx

    def get_source(self, environment, template):
        f = self.ctx.get_file(template)
        source = f.read()
        return source, template, lambda: False


class File(provider.Provider):

    """ Provides file creation using templates or static files. """

    policies = (resources.file.FileApplyPolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(File, self).isvalid(*args, **kwargs)

    def check_path(self, directory, simulate):
        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not os.path.exists(path): #FIXME
                if not simulate:
                    raise error.PathComponentMissing(path)
            elif not os.path.isdir(path):
                raise error.PathComponentNotDirectory(path)

    def has_protected_strings(self):
        def iter(val):
            if isinstance(val, dict):
                for v in val.values():
                    if iter(v):
                        return True
                return False

            elif isinstance(val, list):
                for v in val:
                    if iter(v):
                        return True
                return False

            else:
                return isinstance(val, String)

        return iter(self.resource.template_args)

    def get_template_args(self):
        """ I return a copy of the template_args that contains only basic types (i.e. no protected strings) """
        def _(val):
            if isinstance(val, dict):
                return dict((k,_(v)) for (k,v) in val.items())
            elif isinstance(val, list):
                return list(_(v) for v in val)
            elif isinstance(val, String):
                return val.unprotected
            else:
                return val
        return _(self.resource.template_args)

    def apply(self, context):
        name = self.resource.name

        self.check_path(os.path.dirname(name), context.simulate)

        if self.resource.template:
            # set a special line ending
            # this strips the \n from the template line meaning no blank line,
            # if a template variable is undefined. See ./yaybu/recipe/interfaces.j2 for an example
            env = Environment(loader=YaybuTemplateLoader(context), line_statement_prefix='%')
            template = env.get_template(self.resource.template)
            contents = template.render(self.get_template_args()) + "\n" # yuk
            sensitive = self.has_protected_strings()
        elif self.resource.static:
            contents = context.get_file(self.resource.static).read()
            sensitive = False
        elif self.resource.encrypted:
            contents = context.get_decrypted_file(self.resource.encrypted).read()
            sensitive = True
        else:
            contents = None
            sensitive = False

        # If a file doesn't exist we create an empty one. This means we can
        # ensure the user, group and permissions are in their final state
        # *BEFORE* we write to them.
        created = False
        if not os.path.exists(self.resource.name):
            if not context.simulate:
                with open(self.resource.name, "w") as fp:
                    fp.write("")
                    fp.close()
            created = True

        ac = AttributeChanger(context,
                              self.resource.name,
                              self.resource.owner,
                              self.resource.group,
                              self.resource.mode)
        context.changelog.apply(ac)

        fc = FileContentChanger(context, self.resource.name, contents, sensitive)
        context.changelog.apply(fc)

        if created or fc.changed or ac.changed:
            return True

class RemoveFile(provider.Provider):
    policies = (resources.file.FileRemovePolicy,)

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(RemoveFile, self).isvalid(*args, **kwargs)

    def apply(self, context):
        if os.path.exists(self.resource.name):
            if not os.path.isfile(self.resource.name):
                raise error.InvalidProvider("%r: %s exists and is not a file" % (self, self.resource.name))
            context.shell.execute(["/bin/rm", self.resource.name])
            changed = True
        else:
            context.changelog.debug("File %s missing already so not removed" % self.resource.name)
            changed = False
        return changed


class WatchFile(provider.Provider):
    policies = (resources.file.FileWatchedPolicy, )

    @classmethod
    def isvalid(self, *args, **kwargs):
        return super(WatchFile, self).isvalid(*args, **kwargs)

    def apply(self, context):
        """ Watched files don't have any policy applied to them """
        return self.resource.hash() != self.resource._original_hash




# coding=utf-8

import os
import pwd
import grp
import stat
from yaybu.harness import FakeChrootTestCase


def sibpath(filename):
    return os.path.join(os.path.dirname(__file__), filename)


class TestDirectory(FakeChrootTestCase):

    def test_create_directory(self):
        self.fixture.check_apply("""
            resources:
              - Directory:
                  name: /etc/somedir
                  owner: root
                  group: root
            """)
        self.failUnless(self.fixture.isdir("/etc/somedir"))

    def test_create_directory_and_parents(self):
        self.fixture.check_apply("""
            resources:
                - Directory:
                    name: /etc/foo/bar/baz
                    parents: True
            """)
        self.failUnless(self.fixture.isdir("/etc/foo/bar/baz"))

    def test_remove_directory(self):
        self.fixture.mkdir("/etc/somedir")
        self.fixture.check_apply("""
            resources:
              - Directory:
                  name: /etc/somedir
                  policy: remove
        """)

    def test_remove_directory_recursive(self):
        self.fixture.mkdir("/etc/somedir")
        self.fixture.touch("/etc/somedir/child")
        self.fixture.check_apply("""
            resources:
                - Directory:
                    name: /etc/somedir
                    policy: remove-recursive
            """)
        self.failIfExists("/etc/somedir")

    def test_unicode(self):
        utf8 = "/etc/£££££" # this is utf-8 encoded
        self.fixture.check_apply(open(sibpath("directory_unicode1.yay")).read())
        self.failUnlessExists(utf8)

    def test_attributes(self):
        self.fixture.check_apply("""
            resources:
              - Directory:
                  name: /etc/somedir2
                  owner: nobody
                  group: nogroup
                  mode: 0777
            """)
        self.failUnlessExists("/etc/somedir2")
        st = self.fixture.stat("/etc/somedir2")
        self.failUnless(pwd.getpwuid(st.st_uid)[0] != 'nobody')
        self.failUnless(grp.getgrgid(st.st_gid)[0] != 'nogroup')
        mode = stat.S_IMODE(st.st_mode)
        self.assertEqual(mode, 0777)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
import datetime
from yaybu.core import (resource,
                        argument,
                        policy,
                        error,
                        provider,
                        change,
                        )

from mock import Mock

class F(resource.Resource):
    foo = argument.String(default="42")
    bar = argument.String()

class G(resource.Resource):
    foo = argument.String()
    bar = argument.String()

class H(resource.Resource):
    foo = argument.Integer()
    bar = argument.DateTime()
    baz = argument.File()

class TestResource(unittest.TestCase):

    def test_init(self):
        h = H(**{
            'name': 'test',
            'foo': u'42',
            'bar': u'20100501',
            })
        self.assertEqual(h.foo, 42)
        self.assertEqual(h.bar, datetime.datetime(2010, 05, 01))


class TestArgument(unittest.TestCase):

    def test_storage(self):
        f1 = F(name="test")
        f2 = F(name="test")
        g1 = G(name="test")
        g2 = G(name="test")
        f1.foo = "a"
        f1.bar = "b"
        f2.foo = "c"
        f2.bar = "d"
        g1.foo = "e"
        g1.bar = "f"
        g2.foo = "g"
        g2.bar = "h"
        self.assertEqual(f1.foo, "a")
        self.assertEqual(f1.bar, "b")
        self.assertEqual(f2.foo, "c")
        self.assertEqual(f2.bar, "d")
        self.assertEqual(g1.foo, "e")
        self.assertEqual(g1.bar, "f")
        self.assertEqual(g2.foo, "g")
        self.assertEqual(g2.bar, "h")

    def test_default(self):
        f = F(name="test")
        self.assertEqual(f.foo, "42")

    def test_integer(self):
        h = H(name="test")
        h.foo = u"42"
        self.assertEqual(h.foo, 42)

    def test_datetime(self):
        h = H(name="test")
        h.bar = "20100105"
        self.assertEqual(h.bar, datetime.datetime(2010, 1, 5))


class TestArgumentAssertion(unittest.TestCase):

    def test_present(self):
        class P(policy.Policy):
            signature = [policy.Present("foo")]
        class Q(policy.Policy):
            signature = [policy.Present("bar")]
        class R(policy.Policy):
            signature = [policy.Present("baz")]
        f = F(name="test")
        self.assertEqual(P.conforms(f), True)
        self.assertEqual(Q.conforms(f), False)
        self.assertRaises(AttributeError, R.conforms, f)

    def test_absent(self):
        class P(policy.Policy):
            signature = [policy.Absent("foo")]
        class Q(policy.Policy):
            signature = [policy.Absent("bar")]
        class R(policy.Policy):
            signature = [policy.Absent("baz")]
        f = F(name="test")
        self.assertEqual(P.conforms(f), False)
        self.assertEqual(Q.conforms(f), True)
        self.assertRaises(AttributeError, R.conforms, f)

    def test_and(self):
        class P(policy.Policy):
            signature = [policy.Present("foo"),
                         policy.Absent("bar"),
                         ]
        f = F(name="test")
        self.assertEqual(P.conforms(f), True)

    def test_xor(self):
        class P(policy.Policy):
            signature = [policy.XOR(
                              policy.Present("foo"),
                              policy.Present("bar"),
                         )]
        g = G(name="test")
        self.assertEqual(P.conforms(g), False)
        g.foo = "yes"
        self.assertEqual(P.conforms(g), True)
        g.bar = "yes"
        self.assertEqual(P.conforms(g), False)
        g.foo = None
        self.assertEqual(P.conforms(g), True)


class Ev1(resource.Resource):
    pass

class Ev1FooPolicy(policy.Policy):
    name = "foo"
    resource = Ev1

class Ev1BarPolicy(policy.Policy):
    name = "bar"
    resource = Ev1

class Ev1BazPolicy(policy.Policy):
    name = "baz"
    resource = Ev1

class Ev1Provider(provider.Provider):
    policies = (Ev1FooPolicy, Ev1BarPolicy, Ev1BazPolicy)

    applied = 0

    def apply(self, shell):
        Ev1Provider.applied += 1
        return True

class TestResourceBundle(unittest.TestCase):

    def setUp(self):
        from yaybu.core import event
        event.reset()

    def test_creation(self):
        resources = resource.ResourceBundle.create_from_list([
            {"File": [{
                "name": "/etc/foo",
                "mode": "666",
                }]
             }])
        self.assertEqual(resources["File[/etc/foo]"].mode, 438)

    def test_firing(self):
        Ev1Provider.applied = 0
        resources = resource.ResourceBundle.create_from_list([
            {"Ev1": [
                { "name": "e1",
                  "policy": "foo",
                }, {
                  "name": "e2",
                  "policy":
                      {"baz": [{
                          "when": "foo",
                          "on": "Ev1[e1]",
                          }],
                       },
                  },
            ]}])

        e1 = resources['Ev1[e1]']
        e2 = resources['Ev1[e2]']
        resources.bind()
        self.assertEqual(dict(e2.observers), {})
        self.assertEqual(dict(e1.observers),
                         {'foo': [
                             (True, e2, 'baz')]
                          })
        shell = Mock()
        p1 = e1.get_default_policy().get_provider({})
        p2 = e2.get_default_policy().get_provider({})
        self.assertEqual(p1, Ev1Provider)
        self.assertEqual(p2, provider.NullProvider)
        e1.apply(shell)
        self.assertEqual(Ev1Provider.applied, 1)
        e2.apply(shell)
        self.assertEqual(Ev1Provider.applied, 2)

    def test_not_firing(self):
        Ev1Provider.applied = 0
        resources = resource.ResourceBundle.create_from_list([
            {"Ev1": [
                { "name": "e1",
                  "policy": "foo",
                }, {
                  "name": "e2",
                  "policy":
                      {"baz": [{
                          "when": "baz",
                          "on": "Ev1[e1]",
                          }],
                       },
                  },
            ]}])
        e1 = resources['Ev1[e1]']
        e2 = resources['Ev1[e2]']
        resources.bind()
        self.assertEqual(dict(e2.observers), {})
        self.assertEqual(dict(e1.observers),
                         {'baz': [
                             (True, e2, 'baz')]
                          })
        shell = Mock()
        p1 = e1.get_default_policy().get_provider({})
        p2 = e2.get_default_policy().get_provider({})
        self.assertEqual(p1, Ev1Provider)
        self.assertEqual(p2, provider.NullProvider)
        e1.apply(shell)
        self.assertEqual(Ev1Provider.applied, 1)
        e2.apply(shell)
        self.assertEqual(Ev1Provider.applied, 1)

    def test_forwardreference(self):
        Ev1Provider.applied = 0
        resources = resource.ResourceBundle.create_from_list([
            {"Ev1": [
                { "name": "e1",
                  "policy":
                      {"baz": [{
                          "when": "baz",
                          "on": "Ev1[e2]",
                          }],
                       },
                }, {
                  "name": "e2",
                  "policy": "foo",
                  }
            ]}])
        e1 = resources['Ev1[e1]']
        e2 = resources['Ev1[e2]']
        self.assertRaises(error.BindingError, resources.bind)

    def test_structure(self):
        e1 = Ev1(name="e1",
                policy = {
                    'pol1': {
                        'when': 'bar',
                        'on': 'e2'},
                    })
        e2 = Ev1(name="e2")
        resources = {'e1': e1, 'e2': e2}
        e1.bind(resources)
        e2.bind(resources)
        self.assertEqual(len(e1.observers), 0)
        self.assertEqual(dict(e2.observers), {
            'bar': [(True, e1, 'pol1')]
            })

    def test_multiple(self):
        e1 = Ev1(name="e1",
                policy = {
                    'pol1': [{
                        'when': 'bar',
                        'on': 'e2'}],
                    'pol2': [{
                        'when': 'foo',
                        'on': 'e3'}],
                    'pol3': [{
                        'when': 'baz',
                        'on': 'e2',
                        }]
                    })
        e2 = Ev1(name="e2")
        e3 = Ev1(name="e3")
        resources = {'e1': e1, 'e2': e2, 'e3': e3}
        e1.bind(resources)
        e2.bind(resources)
        self.assertEqual(dict(e1.observers), {})
        self.assertEqual(dict(e2.observers), {
            'bar': [(True, e1, 'pol1')],
            'baz': [(True, e1, 'pol3')],
            })
        self.assertEqual(dict(e3.observers), {
            'foo': [(True, e1, 'pol2')],
            })

    def test_missing(self):
        e1 = Ev1(name="e1",
                policy = {
                    'pol1': [{
                        'when': 'bar',
                        'on': 'missing'}],
                    })
        e2 = Ev1(name="e2")
        resources = {'e1': e1, 'e2': e2}
        self.assertRaises(error.BindingError, e1.bind, resources)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging
import logging.handlers
import subprocess
import getpass

from yaybu.core import error
from yaybu.core import runcontext
from yaybu.core import event

logger = logging.getLogger("runner")

class LoaderError(Exception):
    pass


class Runner(object):

    def trampoline(self, username):
        command = ["sudo", "-u", username] + sys.argv[0:1]

        if "SSH_AUTH_SOCK" in os.environ:
            command.extend(["--ssh-auth-sock", os.environ["SSH_AUTH_SOCK"]])

        command.extend(sys.argv[1:])

        os.execvp(command[0], command)

    def run(self, ctx, bundle=None):
        """ Run locally. """
        if ctx.user and getpass.getuser() != ctx.user:
            self.trampoline(ctx.user)
            return 0

        event.EventState.save_file = "/var/run/yaybu/events.saved"

        # This makes me a little sad inside, but the whole
        # context thing needs a little thought before jumping in
        event.state.simulate = ctx.simulate

        if not ctx.simulate:
            save_parent = os.path.realpath(os.path.join(event.EventState.save_file, os.path.pardir))
            if not os.path.exists(save_parent):
                os.mkdir(save_parent)

        try:
            if os.path.exists(event.EventState.save_file):
                if ctx.resume:
                    event.state.loaded = False
                elif ctx.no_resume:
                    if not ctx.simulate:
                        os.unlink(event.EventState.save_file)
                    event.state.loaded = True
                else:
                    raise error.SavedEventsAndNoInstruction("There is a saved events file - you need to specify --resume or --no-resume")

            # FIXME: We can pull this out to somewhere nicer now
            if bundle:
                ctx.set_bundle(bundle)

            # Actually apply the configuration
            changed = ctx.get_bundle().apply(ctx, ctx.get_config())

            if not ctx.simulate and os.path.exists(event.EventState.save_file):
                os.unlink(event.EventState.save_file)

            if not changed:
                # nothing changed
                ctx.changelog.info("No changes were required")
                return 254

            ctx.changelog.info("All changes were applied successfully")
            return 0

        except error.ExecutionError, e:
            # this will have been reported by the context manager, so we wish to terminate
            # but not to raise it further. Other exceptions should be fully reported with
            # tracebacks etc automatically
            ctx.changelog.error("Terminated due to execution error in processing")
            return e.returncode
        except error.Error, e:
            # If its not an Execution error then it won't have been logged by the
            # Resource.apply() machinery - make sure we log it here.
            ctx.changelog.write(str(e))
            ctx.changelog.error("Terminated due to error in processing")
            return e.returncode
        except SystemExit:
            # A normal sys.exit() is fine..
            raise
        #except:
        #    from yaybu.core.debug import post_mortem
        #    post_mortem()
        #    raise




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, hashlib
from argument import Argument, List, PolicyArgument, String
import policy
import error
import collections
import ordereddict
import event

class ResourceType(type):

    """ Keeps a registry of resources as they are created, and provides some
    simple access to their arguments. """

    resources = {}

    def __new__(meta, class_name, bases, new_attrs):
        cls = type.__new__(meta, class_name, bases, new_attrs)
        cls.__args__ = []
        for b in bases:
            if hasattr(b, "__args__"):
                cls.__args__.extend(b.__args__)
        cls.policies = AvailableResourcePolicies()
        if class_name != 'Resource':
            rname = new_attrs.get("__resource_name__", class_name)
            if rname in meta.resources:
                raise error.ParseError("Redefinition of resource %s" % rname)
            else:
                meta.resources[rname] = cls
        for key, value in new_attrs.items():
            if isinstance(value, Argument):
                cls.__args__.append(key)
        return cls

    @classmethod
    def clear(self):
        self.resources = {}

class AvailableResourcePolicies(dict):

    """ A collection of the policies available for a resource, with some logic
    to work out which of them is the one and only default policy. """

    def default(self):
        default = [p for p in self.values() if p.default]
        if default:
            return default[0]
        else:
            return policy.NullPolicy


class Resource(object):

    """ A resource represents a resource that can be configured on the system.
    This might be as simple as a symlink or as complex as a database schema
    migration. Resources have policies that represent how the resource is to
    be treated. Providers are the implementation of the resource policy.

    Resource definitions specify the complete set of attributes that can be
    configured for a resource. Policies define which attributes must be
    configured for the policy to be used.

    """

    __metaclass__ = ResourceType

    # The arguments for this resource, set in the metaclass
    __args__ = []

    policies = AvailableResourcePolicies()
    """ A dictionary of policy names mapped to policy classes (not objects).

    These are the policies for this resource class.

    Here be metaprogramming magic.

    Dynamically allocated as Yaybu starts up this is effectively static once
    we're up and running. The combination of this attribute and the policy
    argument below is sufficient to determine which provider might be
    appropriate for this resource.

    """

    policy = PolicyArgument()
    """ The list of policies provided by configuration. This is an argument
    like any other, but has a complex representation that holds the conditions
    and options for the policies as specified in the input file. """

    name = String()

    watch = List()
    """ A list of files to monitor while this resource is applied

    The file will be hashed before and after a resource is applied.
    If the hash changes, then it will be like a policy has been applied
    on that file.

    For example::

        resources.append:
          - Execute:
              name: buildout-foobar
              command: buildout2.6
              watch:
                - /var/local/sites/foobar/apache/apache.cfg

          - Service:
              name: apache2
              policy:
                restart:
                  when: watched
                  on: File[/var/local/sites/foobar/apache/apache.cfg]
    """

    def __init__(self, **kwargs):
        """ Pass a dictionary of arguments and they will be updated from the
        supplied data. """
        setattr(self, "name", kwargs["name"])
        for key, value in kwargs.items():
            if not key in self.__args__:
                raise AttributeError("Cannot assign argument '%s' to resource %s" % (key, self))
            setattr(self, key, value)
        self.observers = collections.defaultdict(list)

    def register_observer(self, when, resource, policy, immediately):
        self.observers[when].append((immediately, resource, policy))

    def validate(self, yay=None):
        """ Given the provided yay configuration dictionary, validate that
        this resource is correctly specified. Will raise an exception if it is
        invalid. Returns the provider if it is valid.

        We only validate if:

           - the chosen policies all exist, or
           - there is at least one default policy, and
           - the arguments provided conform with all selected policies, and
           - the selected policies all share a single provider

        If the above is all true then we can identify a provider that should
        be able to implement the required policies.

        """
        if yay is None:
            yay = {}
        this_policy = self.get_default_policy()
        if not this_policy.conforms(self):
            raise error.NonConformingPolicy(this_policy.name)
        # throws an exception if there is not oneandonlyone provider
        provider = this_policy.get_provider(self, yay)
        return True

    def apply(self, context, yay=None, policy=None):
        """ Apply the provider for the selected policy, and then fire any
        events that are being observed. """
        if yay is None:
            yay = {}
        if policy is None:
            pol = self.get_default_policy()
        else:
            pol_class = self.policies[policy]
            pol = pol_class(self)
        prov_class = pol.get_provider(yay)
        prov = prov_class(self)
        changed = prov.apply(context)
        event.state.clear_override(self)
        if changed:
            self.fire_event(pol.name)
        return changed

    def fire_event(self, name):
        """ Apply the appropriate policies on the resources that are observing
        this resource for the firing of a policy. """
        for immediately, resource, policy in  self.observers[name]:
            if immediately is False:
                raise NotImplementedError
            event.state.override(resource, policy)

    def bind(self, resources):
        """ Bind this resource to all the resources on which it triggers.
        Returns a list of the resources to which we are bound. """
        bound = []
        if self.policy is not None:
            for trigger in self.policy.triggers:
                bound.append(trigger.bind(resources, self))
        return bound

    def get_default_policy(self):
        """ Return an instantiated policy for this resource. """
        return event.state.policy(self)

    def dict_args(self):

        """ Return all argument names and values in a dictionary. If an
        argument has no default and has not been set, it's value in the
        dictionary will be None. """

        d = {}
        for a in self.__args__:
            d[a] = getattr(self, a, None)
        return d

    @property
    def id(self):
        classname = getattr(self, '__resource_name__', self.__class__.__name__)
        return "%s[%s]" % (classname, self.name.encode("utf-8"))

    def __repr__(self):
        return self.id

    def __unicode__(self):
        classname = getattr(self, '__resource_name__', self.__class__.__name__)
        return u"%s[%s]" % (classname, self.name)


class ResourceBundle(ordereddict.OrderedDict):

    """ An ordered, indexed collection of resources. Pass in a specification
    that consists of scalars, lists and dictionaries and this class will
    instantiate the appropriate resources into the structure. """

    @classmethod
    def create_from_list(cls, specification):
        """ Given a list of types and parameters, build a resource bundle """
        bundle = cls()
        for spec in specification:
            bundle.add_from_spec(spec)
        return bundle

    @classmethod
    def create_from_yay_expression(cls, expression):
	""" Given a Yay expression that resolves to a list of types and
        parameters, build a resource bundle.  """
        bundle = cls()
        for node in expression:
            spec = node.resolve()
            try:
                bundle.add_from_spec(spec)
            except error.ParseError as exc:
                exc.msg += "\nFile %s, line %d, column %d" % (node.name, node.line, node.column)
                exc.file = node.name
                exc.line = node.line
                exc.column = node.column
                raise

        return bundle

    def key_remap(self, kw):
        """ Maps - to _ to make resource attribute name more pleasant. """
        for k, v in kw.items():
            k = k.replace("-", "_")
            yield str(k),v

    def add_from_spec(self, spec):
        if not hasattr(spec, "keys"):
            raise error.ParseError("Not a valid Resource definition")

        if len(spec.keys()) > 1:
            raise error.ParseError("Too many keys in list item")

        typename, instances = spec.items()[0]
        if not isinstance(instances, list):
            instances = [instances]

        for instance in instances:
            self.add(typename, instance)

    def add(self, typename, instance):
        if not isinstance(instance, dict):
            raise error.ParseError("Expected mapping for %s, got %s" % (typename, instance))

        try:
            kls = ResourceType.resources[typename](**dict(self.key_remap(instance)))
        except KeyError:
            raise error.ParseError("There is no resource type of '%s'" % typename)

        if kls.id in self:
            raise error.ParseError("'%s' cannot be defined multiple times" % kls.id)

        self[kls.id] = kls

        # Create implicit File[] nodes for any watched files
        for watched in instance.get("watch", []):
            w = self.create("File", {
                "name": watched,
                "policy": "watched",
                })
            w._original_hash = w.hash()

        return kls

    # DEPRECATED
    create = add

    def bind(self):
        """ Bind all the resources so they can observe each others for policy
        triggers. """
        for i, resource in enumerate(self.values()):
            for bound in resource.bind(self):
                if bound == resource:
                    raise error.BindingError("Attempt to bind %r to itself!" % resource)
                j = self.values().index(bound)
                if j > i:
                    raise error.BindingError("Attempt to bind forwards on %r" % resource)

    def apply(self, ctx, config):
        """ Apply the resources to the system, using the provided context and
        overall configuration. """
        something_changed = False
        for resource in self.values():
            with ctx.changelog.resource(resource):
                if resource.apply(ctx, config):
                    something_changed = True
        return something_changed

